/**
 * This file is part of cs559ImageProcessing
 * 
 * cs559ImageProcessing is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * cs559ImageProcessing is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with cs559ImageProcessing.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright 2010 James Hill
 */

#include "Filter.h"
#include "ImageSlice.h"
#include "BoundAccess.h"
#include <cmath>

#include <iostream>
using namespace std;

/**
 * This is the default blur function which is used
 * by the majority of the subclassed filters
 */
bool Filter::blur(ImageSlice & input, 
		ImageSlice & output, 
		unsigned int radius,
		BoundAccess & accessor,
		bool renormalize)
{
	// the output will be the same length as the input
	// this function doesn't resample
	output.setLength(input.getLength());

	for(int i = 0; i < input.getLength(); i++)
	{
		output[i].r = output[i].g = output[i].b = 0;

		// start and end at the current sample minus and plus
		// the specified radius
		int sampleStart = i - static_cast<int>(radius);
		int sampleEnd = i + static_cast<int>(radius);
		double weightSum = 0.0;

		// if renormalizing, clamp the start and end to within
		// the image	
		if(renormalize && sampleStart < 0)
			sampleStart = 0;

		if(renormalize && sampleEnd >= input.getLength())
			sampleEnd = input.getLength() - 1;

		// if the radius specified was 1. set the dividing
		// radius equal to the filter radius so that nothing
		// is divided by 0
		double divRadius = radius;
		if(radius == 0.0)
			divRadius = this->radius;

		// run through the sample range and weight each value according to the 
		// calcValue function
		for(int sample = sampleStart; sample <= sampleEnd; sample++)
		{
			// the offset (i-sample) is divided by the (divRadius / filter radius)
			// this brings the sample range into the range of the filter.
			// i.e. the gausian filter goes from -2 to +2 so a sample with
			// radius 10 would be divided by 10/2 or 5 to get 2 which is 
			// withing the filter radius
			double weight = calcValue(static_cast<double>(i - sample) / (divRadius / this->radius));
			output[i].r += accessor.access(sample, input).r * weight;
			output[i].g += accessor.access(sample, input).g * weight;
			output[i].b += accessor.access(sample, input).b * weight;
			weightSum += weight;
		}

		// renormalized the kernel
		output[i].r /= weightSum;
		output[i].g /= weightSum;
		output[i].b /= weightSum;
	}
	
	return true;
}

/**
 * resampling will resize the input to newWidth and store 
 * it in the output slice.
 */
bool Filter::resample(ImageSlice & input, 
			ImageSlice & output, 
			unsigned int newWidth,
			BoundAccess & accessor,
			bool renormalize)
{
	output.setLength(newWidth);

	// the algorithm steps through the rang of the newWidth
	// but the original input needs to be referenced.  The 
	// step value represents how much one unit of the newWidth
	// represents in the old width		
	double step = static_cast<double>(input.getLength() - 1) / static_cast<double>(newWidth - 1);

	// depending on if we are upsampling or downsampling, the 
	// value we use to determine the sample range and the 
	// scaling of the calcValue sample will be either 1 or step
	double divisor = 1.0;
	if(newWidth < input.getLength())
		divisor = step;

	// iterate through each value of the new ImageSlice and 
	// keep track of the corresponding value in the original input
	// with currentStep
	double currentStep = 0;
	for(int i = 0; i < newWidth; i++, currentStep += step)
	{
		// calculate what values of the orignal signal will
		// fall withing the radius of the filter
		int sampleStart = static_cast<int>(ceil(currentStep - radius * divisor));
		int sampleEnd = static_cast<int>(floor(currentStep + radius * divisor));
		double weightSum = 0;
		output[i].r = output[i].g = output[i].b = 0;

		// if renormalizing, clamp the start and end signal
		// indicies to 0 and the input length	
		if(renormalize && sampleStart < 0)
			sampleStart = 0;

		if(renormalize && sampleEnd >= input.getLength())
			sampleEnd = input.getLength() - 1;

		// calculate, weight, and sum over the sample range.	
		for(int sample = sampleStart; sample <= sampleEnd; sample++)
		{
			double weight = calcValue((static_cast<double>(sample) - currentStep)/divisor);
			output[i].r += accessor.access(sample, input).r * weight;
			output[i].g += accessor.access(sample, input).g * weight;
			output[i].b += accessor.access(sample, input).b * weight;
			weightSum += weight;
		}
		output[i].r /= weightSum;
		output[i].g /= weightSum;
		output[i].b /= weightSum;
	}
	return true;
}
		
