/**
 * This file is part of cs559ImageProcessing
 * 
 * cs559ImageProcessing is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * cs559ImageProcessing is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with cs559ImageProcessing.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright 2010 James Hill
 */

#include "Image.h"
#include "RGBValue.h"
#include "TargaImage.h"
#include "ImageSlice.h"

#include <iostream>
using namespace std;

/**
 * Constructor, build a new with dimentions width and height
 */		
Image::Image(unsigned int width, unsigned int height) : width(width), height(height)
{
	if(width == 0 && height == 0)
		values = 0;
	else
		values = new RGBValue[width * height];
}

/**
 * destructor
 */
Image::~Image()
{
	if(values)
		delete(values);
}

/**
 * Loads an image from a file using the TargaImage class
 */
bool Image::load(const string & file)
{
	bool returnValue = true;

	TargaImage * loadedTarga = TargaImage::readImage(file.c_str());	
	if(loadedTarga == 0)
		return false;

	if(!(this->load(*loadedTarga)))
		returnValue = false;
	
	delete loadedTarga;

	return returnValue;	
}

/**
 * writes and image to a file using the TargaImage
 * class
 */
bool Image::write(const string & file)
{
	bool returnValue = true;

	TargaImage * convertedTarga = this->buildTarga();
	if(convertedTarga == 0)
		return false;

	if(convertedTarga->write(file.c_str()) != 1)
		returnValue = false;

	delete convertedTarga;

	return returnValue;
}

/**
 * Loads and image from a TargaImage class.  This
 * function converts all unsigned char's into doubles
 */
bool Image::load(const TargaImage & image)
{
	if(values)
		delete values;

	width = image.width();
	height = image.height();
	values = new RGBValue[width * height];
	
	unsigned char * pixels = image.pixels();

	for(unsigned int i = 0, stop = width * height; i < stop; i++, pixels += 4)
	{
		values[i].setR(pixels[0]);	
		values[i].setG(pixels[1]);	
		values[i].setB(pixels[2]);	
		
	}

	return true;
}

/**
 * builds a TargaImage class using the data in the
 * image.  This function converts the double
 * values back into unsigned char values
 */
TargaImage * Image::buildTarga()
{
	TargaImage * newImage = TargaImage::blankImage(width, height);

	if(newImage == 0)
		return 0;

	unsigned char * pixels = newImage->pixels();
	
	for(unsigned int i = 0, stop = width * height; i < stop; i++, pixels+=4)
	{
		pixels[0] = values[i].getR();	
		pixels[1] = values[i].getG();	
		pixels[2] = values[i].getB();	
		pixels[3] = 0xff;
	}
	
	return newImage;
}

/**
 * returns a reference to the value at x, y.  
 * This is not where bound access is determined
 */	
const RGBValue & Image::get(unsigned int x, unsigned int y) const
{
	return values[y*width+x];
}

/**
 * like the above but not const
 */
RGBValue & Image::get(unsigned int x, unsigned int y)
{
	return values[y*width+x];
}

/**
 * returns an ImageSlice of the row
 */		
bool Image::getRow(unsigned int row, ImageSlice & out)
{
	if(row >= height)
		return false;

	out.setLength(width);
	for(unsigned int i = 0; i < width; i++)
		out[i] = get(i, row);

	return true;	
}

/**
 * inserts the ImageSlice into the image at row
 */
bool Image::putRow(unsigned int row, const ImageSlice & in)
{
	if(row >= height || in.getLength() != width)
		return false;

	for(unsigned int i = 0; i < width; i++)
		get(i, row) = in[i];

	return true;
}

/**
 * returns an ImageSlice of the col
 */	
bool Image::getCol(unsigned int col, ImageSlice & out)
{
	if(col >= width)
		return false;
	
	out.setLength(height);
	for(unsigned int i = 0; i < height; i++)
		out[i] = get(col, i);

	return true;
}

/**
 * inserts the ImageSlice into the image at col
 */
bool Image::putCol(unsigned int col, const ImageSlice & in)
{
	if(col >= width || in.getLength() != height)
		return false;

	for(unsigned int i = 0; i < height; i++)
		get(col, i) = in[i];

	return true;
}

