/**
 * This file is part of cs559ImageProcessing
 * 
 * cs559ImageProcessing is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * cs559ImageProcessing is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with cs559ImageProcessing.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright 2010 James Hill
 */

#include <iostream>
using namespace std;

#include "CommandLineUtils.h"
#include "BoundTypes.h"
#include "FilterTypes.h"
#include "Image.h"
#include "libtarga.h"
#include "ImageOperations.h"
#include "Filter.h"

void printUsage();
bool loadParameters(int argc, char ** argv, ArgList & args);

int main(int argc, char** argv)
{
	// args input output radius [filter]
	ArgList args;
	if(!loadParameters(argc, argv, args))
		return -1;

	// load the input image
	Image input;
	if(!input.load(args.input))
	{
		cout << "ERROR: Unable to load [" << args.input << "] : " << tga_error_string(tga_get_last_error()) << endl;
		return -1;
	}

	// set up the access policy for the image
	BoundType boundType = static_cast<BoundType>(args.params[2]);
	BoundAccess * boundAccess = getBoundType(boundType);

	// create the specified kernel
	Filter * filter = getFilter(static_cast<FilterType>(args.params[1]));

	// set up the output kernel
	Image output(input.getWidth(), input.getHeight());

	cout << "Image:  " << input.getWidth() << "x" << input.getHeight() << endl;
	cout << "Radius: " << args.params[0] << endl;
	cout << "Kernel: " << filterToString(static_cast<FilterType>(args.params[1])) << endl;
	cout << "Bound:  " << boundTypeToString(static_cast<BoundType>(args.params[2])) << endl;

	// filter the input into the output
	// renormalize if the kernel goes outside the image
	// don't allow the filter to resize the kernel	
	blurImage(input, output, *boundAccess, *filter, args.params[0], boundType == RENORMALIZE);

	// write the image
	if(!output.write(args.output))
	{
		cout << "ERROR: Unable to write file [" << args.output << "]" << endl;
		return -1;	
	}	
	
	return 0;
}

void printUsage()
{
	cout << endl;
	cout << "\tblur input output radius [filter] [bound]" << endl;
	cout << endl;
	cout << "\t\tinput  - File name of input image" << endl;
	cout << "\t\toutput - File name of output image" << endl;
	cout << "\t\tradius - Positive integer value specifying" << endl;
	cout << "\t\t         how many pixels on each side of a" << endl;
	cout << "\t\t         given pixel the filter should be" << endl;
	cout << "\t\t         applied" << endl;
	cout << "\t\tfilter - An optional argument defaulting to" << endl;
	cout << "\t\t         2 (Tent).  The following are options:" << endl;
	cout << endl;
	cout << "\t\t         0 - Spike Kernel" << endl;
	cout << "\t\t         1 - Box Kernel" << endl;
	cout << "\t\t         2 - Tent Kernel" << endl;
	cout << "\t\t         3 - Gaussian Kernel" << endl;
	cout << "\t\t         4 - B-Spline Cubic" << endl;
	cout << endl;
	cout << "\t\tbound  - How boundries are handled." << endl;
	cout << "\t\t         defaults to 3 (renormailze)" << endl;
	cout << "\t\t         options are:" << endl;
	cout << endl;
	cout << "\t\t         0 - zero" << endl;
	cout << "\t\t         1 - clamp" << endl;
	cout << "\t\t         2 - mirror" << endl;
	cout << "\t\t         3 - renormalize" << endl;
	cout << endl;	
}

bool loadParameters(int argc, char ** argv, ArgList & args)
{
	// make sure we have enough parameters
	if(argc < 4)
	{
		cout << "ERROR: Invalid parameter count" << endl;
		printUsage();
		return false;
	}

	// load in the arguments and make sure that they are valid	
	int error = loadArguments(argc, argv, args);
	if(error)
	{
		cout << "ERROR: Invalid parameter [" << argv[error] << "]" << endl;
		printUsage();
		return false;
	}

	// if we have no parameters then there was no radius specified
	if(args.params.size() == 0)
	{
		cout << "ERROR: Radius not specified" << endl;
		printUsage();
		return false;
	}

	// if there are more than two int params, warn the user that
	// too many were specified
	if(args.params.size() > 3)
		cout << "WARNING: Specified too many arguments" << endl;

	if(args.input == args.output)
		cout << "WARNING: Input and output image names are the same" << endl;

	// if there is a kernel type specified, make sure it's valid
	if(args.params.size() >= 2)
	{
		if(!isValidFilter(static_cast<FilterType>(args.params[1])))
		{
			cout << "ERROR: Invalid filter type [" << args.params[1] << "]" << endl;
			printUsage();
			return false;
		}
	}
	else // otherwise default to 1
	{
		args.params.push_back(2);
	}

	// check for bounds option
	if(args.params.size() >= 3)
	{
		if(!isValidBoundType(static_cast<BoundType>(args.params[2])))
		{
			cout << "ERROR: Invalid bound value [" << args.params[2] << "]" << endl;
			printUsage();
			return false;
		}
	}
	else
	{
		args.params.push_back(3);
	}

	// everything looks good, return true;
	return true;
}

